import datetime,pkg_resources

name = "miner2"
GIT_SHA = '$Id: 563821013b1f4189d012b54416a7989396d0811d $'

try:
    __version__ = pkg_resources.get_distribution(name)
except:
    __version__ = 'development'

message=str(__version__).replace('miner2','miner2 version')
#print(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S \t hello from {}".format(message)))



import numpy, scipy, pandas, sklearn, lifelines, matplotlib, seaborn

import multiprocessing, multiprocessing.pool
import sys


def write_dependency_infos(outfile):
    py_version = map(str, list(sys.version_info[0:3]))
    outfile.write('Python %s\n\n' % ('.'.join(py_version)))
    outfile.write('Dependencies:\n')
    outfile.write('-------------\n\n')
    outfile.write('numpy: %s\n' % numpy.__version__)
    outfile.write('scipy: %s\n' % scipy.__version__)
    outfile.write('pandas: %s\n' % scipy.__version__)
    outfile.write('sklearn: %s\n' % sklearn.__version__)
    outfile.write('lifelines: %s\n' % lifelines.__version__)
    outfile.write('matplotlib: %s\n' % matplotlib.__version__)
    outfile.write('seaborn: %s\n' % seaborn.__version__)


def split_for_multiprocessing(vector, cores):
    partition = int(len(vector) / cores)
    remainder = len(vector) - cores * partition
    starts = numpy.arange(0, len(vector) ,partition)[0:cores]

    for i in range(remainder):
        starts[cores - remainder + i] = starts[cores - remainder + i] + i

    stops = starts + partition
    for i in range(remainder):
        stops[cores - remainder + i] = stops[cores - remainder + i] + 1

    return list(zip(starts, stops))


def multiprocess(function, tasks):
    hydra = multiprocessing.pool.Pool(len(tasks))
    output = hydra.map(function, tasks)
    hydra.close()
    hydra.join()
    return output


def condense_output(output):
    """WW: what is this weird construct doing ???? It seems like it does things very complicated"""
    results = {}
    for i in range(len(output)):
        resultsDict = output[i]
        keys = list(resultsDict.keys())
        for j in range(len(resultsDict)):
            key = keys[j]
            results[key] = resultsDict[key]
    return results



import numpy
import pandas
from scipy import stats
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import sklearn
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
import time

import traceback as tb
import logging


"""
Functions used for inferring sample subtypes
"""

def f1_decomposition(sampleMembers=None, thresholdSFM=0.333, sampleFrequencyMatrix=None):
    # thresholdSFM is the probability cutoff that makes the density of the binary
    # similarityMatrix = 0.15
    # sampleMembers is a dictionary with features as keys and members as elements
    # sampleFrequencyMatrix[i,j] gives the probability that sample j appears in a
    # cluster given that sample i appears
    if sampleFrequencyMatrix is None:
        sampleFrequencyMatrix = sample_coincidence_matrix(sampleMembers, freqThreshold=thresholdSFM,
                                                          frequencies=True)
    # similarityMatrix is defined such that similarityMatrix[i,j] = 1 iff
    # sampleFrequencyMatrix[i,j] >= thresholdSFM
    similarityMatrix = sampleFrequencyMatrix * sampleFrequencyMatrix.T
    similarityMatrix[similarityMatrix < thresholdSFM] = 0
    similarityMatrix[similarityMatrix > 0] = 1

    # remainingMembers is the set of set of unclustered members
    remainingMembers = set(similarityMatrix.index)

    # probeSample is the sample that serves as a seed to identify a cluster in a given iteration
    # WW: replaced deprecated numpy function with idxmax() on data frame
    probeSample = similarityMatrix.sum(axis=1).idxmax()

    # members are the samples that satisfy the similarity condition with the previous cluster or probeSample
    members = set(similarityMatrix.index[numpy.where(similarityMatrix[probeSample] == 1)[0]])
    # nonMembers are the remaining members not in the current cluster
    nonMembers = remainingMembers-members
    # instantiate list to collect clusters of similar members
    similarityClusters = []
    # instantiate f1 score for optimization
    f1 = 0

    for iteration in range(1500):

        predictedMembers = members
        predictedNonMembers = remainingMembers-predictedMembers

        sumSlice = numpy.sum(similarityMatrix.loc[:,list(predictedMembers)], axis=1) / float(len(predictedMembers))
        members = set(similarityMatrix.index[numpy.where(sumSlice > 0.8)[0]])

        if members == predictedMembers:
            similarityClusters.append(list(predictedMembers))
            if len(predictedNonMembers) == 0:
                break
            similarityMatrix = similarityMatrix.loc[predictedNonMembers, predictedNonMembers]

            # replaced deprecated numpy function with DataFrame.idxmax()
            probeSample = similarityMatrix.sum(axis=1).idxmax()

            members = set(similarityMatrix.index[numpy.where(similarityMatrix[probeSample]==1)[0]])
            remainingMembers = predictedNonMembers
            nonMembers = remainingMembers-members
            f1 = 0
            continue

        nonMembers = remainingMembers-members
        TP = len(members&predictedMembers)
        FN = len(predictedNonMembers & members)
        FP = len(predictedMembers & nonMembers)
        tmpf1 = TP/float(TP+FN+FP)

        if tmpf1 <= f1:
            similarityClusters.append(list(predictedMembers))
            if len(predictedNonMembers) == 0:
                break
            similarityMatrix = similarityMatrix.loc[predictedNonMembers,predictedNonMembers]

            # replaced deprecated numpy function with DataFrame.idxmax()
            probeSample = similarityMatrix.sum(axis=1).idxmax()

            members = set(similarityMatrix.index[numpy.where(similarityMatrix[probeSample]==1)[0]])
            remainingMembers = predictedNonMembers
            nonMembers = remainingMembers-members
            f1 = 0
            continue

        elif tmpf1 > f1:
            f1 = tmpf1
            continue

    similarityClusters.sort(key = lambda s: -len(s))
    return similarityClusters


def sample_coincidence_matrix(dict_, freqThreshold=0.333, frequencies=False):

    keys = dict_.keys()
    lists = [dict_[key] for key in keys]
    samples = list(set(numpy.hstack(lists)))

    template = pandas.DataFrame(numpy.zeros((len(samples),len(samples))))
    template.index = samples
    template.columns = samples
    for key in keys:
        hits = dict_[key]
        template.loc[hits,hits]+=1
    trace = numpy.array([template.iloc[i,i] for i in range(template.shape[0])])
    normDf = ((template.T)/trace).T
    if frequencies is not False:
        return normDf
    normDf[normDf<freqThreshold]=0
    normDf[normDf>0]=1

    return normDf


def plot_similarity(similarityMatrix, orderedSamples, vmin=0, vmax=0.5,
                    title="Similarity matrix", xlabel="Samples", ylabel="Samples",
                    fontsize=14, figsize=(7,7), savefig=None):
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    try:
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    except:
        tb.print_exc()  # don't silence exceptions

    ax.imshow(similarityMatrix.loc[orderedSamples,orderedSamples],cmap='viridis',vmin=vmin,vmax=vmax)
    ax.grid(False)
    plt.title(title,FontSize=fontsize+2)
    plt.xlabel(xlabel,FontSize=fontsize)
    plt.ylabel(ylabel,FontSize=fontsize)
    if savefig is not None:
        plt.savefig(savefig,bbox_inches="tight")
    return


def centroid_expansion(classes, sampleMatrix, f1Threshold=0.3, returnCentroids=None):
    print('centroid_expansion')
    centroids = []
    for i in range(len(classes)):
        clusterComponents = sampleMatrix.loc[:,classes[i]]
        class1 = numpy.mean(clusterComponents, axis=1)
        hits = numpy.where(class1 > 0.6)[0]
        centroid = pandas.DataFrame(sampleMatrix.iloc[:, 0])
        centroid.columns = [i]
        centroid[i] = 0
        centroid.iloc[hits, 0] = 1
        centroids.append(centroid)

    miss = []
    centroidClusters = [[] for i in range(len(centroids))]
    for smpl in sampleMatrix.columns:
        probeVector = numpy.array(sampleMatrix[smpl])
        scores = []
        for ix in range(len(centroids)):
            tmp = __f1(numpy.array(probeVector), centroids[ix])
            scores.append(tmp)
        scores = numpy.array(scores)
        match = numpy.argmax(scores)
        if scores[match] < f1Threshold:
            miss.append(smpl)
        elif scores[match] >= f1Threshold:
            centroidClusters[match].append(smpl)

    centroidClusters.append(miss)

    if returnCentroids is not None:
        centroidMatrix = pandas.DataFrame(pandas.concat(centroids, axis=1))
        return centroidClusters, centroidMatrix

    return centroidClusters

def get_centroids(classes, sampleMatrix):
    centroids = []
    for i in range(len(classes)):
        clusterComponents = sampleMatrix.loc[:,classes[i]]
        class1 = np.mean(clusterComponents,axis=1)
        centroid = pd.DataFrame(class1)
        centroid.columns = [i]
        centroid.index = sampleMatrix.index
        centroids.append(centroid)
    return pd.concat(centroids,axis=1)



def __f1(vector1, vector2):
    members = set(numpy.where(vector1 == 1)[0])
    nonMembers = set(numpy.where(vector1 == 0)[0])
    predictedMembers = set(numpy.where(vector2 == 1)[0])
    predictedNonMembers = set(numpy.where(vector2 == 0)[0])

    TP = len(members & predictedMembers)
    FN = len(predictedNonMembers & members)
    FP = len(predictedMembers & nonMembers)
    if TP == 0:
        return 0.0
    F1 = TP / float(TP + FN + FP)

    return F1


def map_expression_to_network(centroidMatrix, membershipMatrix, threshold=0.05):
    logging.info('map_expression_to_network()')
    miss = []
    centroidClusters = [[] for i in range(centroidMatrix.shape[1])]
    for smpl in membershipMatrix.columns:
        probeVector = numpy.array(membershipMatrix[smpl])
        scores = []
        for ix in range(centroidMatrix.shape[1]):
            tmp = __f1(numpy.array(probeVector), numpy.array(centroidMatrix.iloc[:, ix]))
            scores.append(tmp)

        scores = numpy.array(scores)
        match = numpy.argmax(scores)

        if scores[match] < threshold:
            miss.append(smpl)
        elif scores[match] >= threshold:
            centroidClusters[match].append(smpl)

    centroidClusters.append(miss)
    return centroidClusters


def order_membership(centroidMatrix, membershipMatrix, mappedClusters, ylabel="",
                     resultsDirectory=None, showplot=False):
    centroidRank = []
    alreadyMapped = []
    for ix in range(centroidMatrix.shape[1]):
        tmp = numpy.where(centroidMatrix.iloc[:,ix] == 1)[0]
        signature = list(set(tmp) - set(alreadyMapped))
        centroidRank.extend(signature)
        alreadyMapped.extend(signature)

    orderedClusters = centroidMatrix.index[numpy.array(centroidRank)]
    try:
        ordered_matrix = membershipMatrix.loc[orderedClusters, numpy.hstack(mappedClusters)]
    except:
        tb.print_exc()  # don't silence exceptions
        ordered_matrix = membershipMatrix.loc[numpy.array(orderedClusters).astype(int),
                                              numpy.hstack(mappedClusters)]

    if showplot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        try:
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        except:
            tb.print_exc()  # don't silence exceptions

        ax.imshow(ordered_matrix,cmap='viridis',aspect="auto")
        ax.grid(False)

        plt.title(ylabel.split("s")[0]+"Activation",FontSize=16)
        plt.xlabel("Samples",FontSize=14)
        plt.ylabel(ylabel,FontSize=14)
        if resultsDirectory is not None:
            plt.savefig(os.path.join(resultsDirectory,"binaryActivityMap.pdf"))

    return ordered_matrix

def plot_differential_matrix(overExpressedMembersMatrix, underExpressedMembersMatrix,
                             orderedOverExpressedMembers, cmap="viridis", aspect="auto",
                             saveFile=None, showplot=False):
    differentialActivationMatrix = overExpressedMembersMatrix - underExpressedMembersMatrix
    orderedDM = differentialActivationMatrix.loc[orderedOverExpressedMembers.index,
                                                 orderedOverExpressedMembers.columns]

    if showplot:
        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(111)
        try:
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        except:
            tb.print_exc()  # don't silence exceptions

        ax.imshow(orderedDM,cmap=cmap,vmin=-1,vmax=1,aspect=aspect)
        ax.grid(False)
        if saveFile is not None:
            plt.ylabel("Regulons",FontSize=14)
            plt.xlabel("Samples",FontSize=14)
            ax.grid(False)
            plt.savefig(saveFile,bbox_inches="tight")

    return orderedDM


def mosaic(dfr, clusterList, minClusterSize_x=4, minClusterSize_y=5, allow_singletons=True,
           max_groups=50, saveFile=None, random_state=12):
    print('mosaic()')
    lowResolutionPrograms = [[] for i in range(len(clusterList))]
    sorting_hat = []
    for i in range(len(clusterList)):
        patients = clusterList[i]
        if len(patients) < minClusterSize_x:
            continue
        subset = dfr.loc[:,patients]
        density = subset.sum(axis=1)/float(subset.shape[1])
        sorting_hat.append(numpy.array(density))

    enrichment_matrix = numpy.vstack(sorting_hat).T
    choice = numpy.argmax(enrichment_matrix,axis=1)
    for i in range(dfr.shape[0]):
        lowResolutionPrograms[choice[i]].append(dfr.index[i])

    # Cluster modules into transcriptional programs
    y_clusters = []
    for program in range(len(lowResolutionPrograms)):
        regs = lowResolutionPrograms[program]
        if len(regs) == 0:
            continue
        df = dfr.loc[regs,:]
        sil_scores = []
        max_clusters_y = min(max_groups,int(len(regs)/3.))
        for numClusters_y in range(2,max_clusters_y):
            clusters_y, labels_y, centroids_y = kmeans(df,numClusters=numClusters_y,
                                                       random_state=random_state)
            lens_y = [len(c) for c in clusters_y]
            if min(lens_y) < minClusterSize_y:
                if allow_singletons is True:
                    if min(lens_y) != 1:
                        kmSS = 0
                        sil_scores.append(kmSS)
                        continue
                    elif min(lens_y) == 1:
                        pass
                elif allow_singletons is not True:
                    kmSS = 0
                    sil_scores.append(kmSS)
                    continue

            clusters_y.sort(key=lambda s: -len(s))
            kmSS=sklearn.metrics.silhouette_score(df,labels_y,metric='euclidean')
            sil_scores.append(kmSS)

        if len(sil_scores) > 0:
            top_hit = min(numpy.where(numpy.array(sil_scores)>=0.95*max(sil_scores))[0] + 2)
            clusters_y, labels_y, centroids_y = kmeans(df, numClusters=top_hit,
                                                       random_state=random_state)
            clusters_y.sort(key=lambda s: -len(s))
            y_clusters.append(list(clusters_y))

        elif len(sil_scores) == 0:
            y_clusters.append(regs)

    order_y = numpy.hstack([numpy.hstack(y_clusters[i]) for i in range(len(y_clusters))])

    #Cluster patients into subtype states
    x_clusters = []
    for c in range(len(clusterList)):
        patients = clusterList[c]
        if len(patients)<= minClusterSize_x:
            x_clusters.append(patients)
            continue

        if allow_singletons is not True:
            if len(patients)<= 2*minClusterSize_x:
                x_clusters.append(patients)
                continue

        if len(patients) == 0:
            continue
        df = dfr.loc[order_y,patients].T
        sil_scores = []

        max_clusters_x = min(max_groups,int(len(patients)/3.))
        for numClusters_x in range(2,max_clusters_x):
            clusters_x, labels_x, centroids_x = kmeans(df, numClusters=numClusters_x,
                                                       random_state=random_state)
            lens_x = [len(c) for c in clusters_x]
            if min(lens_x) < minClusterSize_x:
                if allow_singletons is True:
                    if min(lens_x) != 1:
                        kmSS = 0
                        sil_scores.append(kmSS)
                        continue
                    elif min(lens_x) == 1:
                        pass
                elif allow_singletons is not True:
                    kmSS = 0
                    sil_scores.append(kmSS)
                    continue

            clusters_x.sort(key=lambda s: -len(s))

            kmSS=sklearn.metrics.silhouette_score(df,labels_x,metric='euclidean')
            sil_scores.append(kmSS)

        if len(sil_scores) > 0:
            top_hit = min(numpy.where(numpy.array(sil_scores) >= 0.999 * max(sil_scores))[0] + 2)
            clusters_x, labels_x, centroids_x = kmeans(df, numClusters=top_hit,
                                                       random_state=random_state)
            clusters_x.sort(key=lambda s: -len(s))
            x_clusters.append(list(clusters_x))
        elif len(sil_scores) == 0:
            x_clusters.append(patients)
    try:
        micro_states = []
        for i in range(len(x_clusters)):
            if len(x_clusters[i])>0:
                if type(x_clusters[i][0]) is not str:
                    for j in range(len(x_clusters[i])):
                        micro_states.append(x_clusters[i][j])
                elif type(x_clusters[i][0]) is str:
                    micro_states.append(x_clusters[i])

        order_x = numpy.hstack(micro_states)
        fig = plt.figure(figsize=(7,7))
        ax = fig.gca()
        ax.imshow(dfr.loc[order_y,order_x],cmap="bwr",vmin=-1,vmax=1)
        ax.set_aspect(dfr.shape[1]/float(dfr.shape[0]))
        ax.grid(False)
        ax.set_ylabel("Regulons",FontSize=14)
        ax.set_xlabel("Samples",FontSize=14)
        if saveFile is not None:
            plt.savefig(saveFile,bbox_inches="tight")

        return y_clusters, micro_states
    except:
        tb.print_exc()  # don't silence exceptions

    return y_clusters, x_clusters


def kmeans(df, numClusters, random_state=None):

    if random_state is not None:
        # Number of clusters
        kmeans = KMeans(n_clusters=numClusters,random_state=random_state)

    elif random_state is None:
        # Number of clusters
        kmeans = KMeans(n_clusters=numClusters)

    # Fitting the input data
    kmeans = kmeans.fit(df)
    # Getting the cluster labels
    labels = kmeans.predict(df)
    # Centroid values
    centroids = kmeans.cluster_centers_

    clusters = []
    for i in range(numClusters):
        clstr = df.index[numpy.where(labels==i)[0]]
        clusters.append(clstr)

    return clusters, labels, centroids


def transcriptional_programs(programs, reference_dictionary):
    transcriptionalPrograms = {}
    programRegulons = {}
    p_stack = []
    programs_flattened = numpy.array(programs).flatten()
    for i in range(len(programs_flattened)):
        if len(numpy.hstack(programs_flattened[i])) > len(programs_flattened[i]):
            for j in range(len(programs_flattened[i])):
                p_stack.append(list(programs_flattened[i][j]))
        else:
            p_stack.append(list(programs_flattened[i]))

    for j in range(len(p_stack)):
        key = ("").join(["TP",str(j)])
        regulonList = [i for i in p_stack[j]]
        programRegulons[key] = regulonList
        tmp = [reference_dictionary[i] for i in p_stack[j]]
        transcriptionalPrograms[key] = list(set(numpy.hstack(tmp)))
    return transcriptionalPrograms, programRegulons


# =============================================================================
# Functions used for cluster analysis
# =============================================================================

def get_eigengenes(coexpressionModules, expressionData, regulon_dict=None, saveFolder=None):
    eigengenes = principal_df(coexpressionModules, expressionData, subkey=None,
                              regulons=regulon_dict, minNumberGenes=1)
    eigengenes = eigengenes.T
    index = numpy.sort(numpy.array(eigengenes.index).astype(int))
    eigengenes = eigengenes.loc[index.astype(str),:]
    if saveFolder is not None:
        eigengenes.to_csv(os.path.join(saveFolder,"eigengenes.csv"))
    return eigengenes


def __regulon_dictionary(regulons):
    regulonModules = {}
    #str(i):[regulons[key][j]]}
    df_list = []

    for tf in regulons.keys():
        for key in regulons[tf].keys():
            genes = regulons[tf][key]
            id_ = str(len(regulonModules))
            regulonModules[id_] = regulons[tf][key]
            for gene in genes:
                df_list.append([id_,tf,gene])

    array = numpy.vstack(df_list)
    df = pandas.DataFrame(array)
    df.columns = ["Regulon_ID","Regulator","Gene"]

    return regulonModules, df


def principal_df(dict_, expressionData, regulons=None, subkey='genes',
                   minNumberGenes=8, random_state=12):

    pcDfs = []
    setIndex = set(expressionData.index)

    if regulons is not None:
        dict_, df = __regulon_dictionary(regulons)
    for i in dict_.keys():
        if subkey is not None:
            genes = list(set(dict_[i][subkey])&setIndex)
            if len(genes) < minNumberGenes:
                continue
        elif subkey is None:
            genes = list(set(dict_[i])&setIndex)
            if len(genes) < minNumberGenes:
                continue

        pca = PCA(1,random_state=random_state)
        principalComponents = pca.fit_transform(expressionData.loc[genes,:].T)
        principalDf = pandas.DataFrame(principalComponents)
        principalDf.index = expressionData.columns
        principalDf.columns = [str(i)]

        normPC = numpy.linalg.norm(numpy.array(principalDf.iloc[:,0]))
        pearson = stats.pearsonr(principalDf.iloc[:,0],
                                 numpy.median(expressionData.loc[genes,:],
                                              axis=0))
        signCorrection = pearson[0] / numpy.abs(pearson[0])

        principalDf = signCorrection*principalDf / normPC

        pcDfs.append(principalDf)

    principalMatrix = pandas.concat(pcDfs,axis=1)

    return principalMatrix


def reduce_modules(df, programs, states, stateThreshold=0.75, saveFile=None):

    df = df.loc[:, numpy.hstack(states)]
    statesDf = pandas.DataFrame(numpy.zeros((len(programs), df.shape[1])))
    statesDf.index = range(len(programs))
    statesDf.columns = df.columns

    for i in range(len(programs)):
        state = programs[i]
        subset = df.loc[state, :]

        state_scores = subset.sum(axis=0) / float(subset.shape[0])

        keep_high = numpy.where(state_scores >= stateThreshold)[0]
        keep_low = numpy.where(state_scores <= -1 * stateThreshold)[0]
        hits_high = numpy.array(df.columns)[keep_high]
        hits_low = numpy.array(df.columns)[keep_low]

        statesDf.loc[i, hits_high] = 1
        statesDf.loc[i, hits_low] = -1

    if saveFile is not None:
        fig = plt.figure(figsize=(7,7))
        ax = fig.gca()
        ax.imshow(statesDf, cmap="bwr", vmin=-1, vmax=1, aspect='auto')
        ax.grid(False)
        ax.set_ylabel("Transcriptional programs", FontSize=14)
        ax.set_xlabel("Samples", FontSize=14)
        plt.savefig(saveFile, bbox_inches="tight")

    return statesDf


def programs_vs_states(statesDf, states, filename=None, showplot=False):
    pixel = numpy.zeros((statesDf.shape[0], len(states)))
    for i in range(statesDf.shape[0]):
        for j in range(len(states)):
            pixel[i,j] = numpy.mean(statesDf.loc[statesDf.index[i],states[j]])

    pixel = pandas.DataFrame(pixel)
    pixel.index = statesDf.index

    if showplot is False:
        return pixel

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.imshow(pixel,cmap="bwr",vmin=-1,vmax=1,aspect="auto")
    ax.grid(False)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    plt.ylabel("Transcriptional programs",FontSize=14)
    plt.xlabel("Transcriptional states",FontSize=14)
    if filename is not None:
        plt.savefig(filename,bbox_inches="tight")

    return pixel


def tsne(matrix, perplexity=100, n_components=2, n_iter=1000, plotOnly=True, plotColor="red",
         alpha=0.4, dataOnly=False):
    X = numpy.array(matrix.T)
    X_embedded = TSNE(n_components=n_components, n_iter=n_iter, n_iter_without_progress=300,
                      init='random', random_state=0, perplexity=perplexity).fit_transform(X)
    if plotOnly is True:
        plt.scatter(X_embedded[:,0],X_embedded[:,1],color=plotColor,alpha=alpha)
        return
    if dataOnly is True:
        return X_embedded

    plt.scatter(X_embedded[:,0],X_embedded[:,1],color=plotColor,alpha=alpha)

    return X_embedded


def tsne_state_labels(tsneDf, states):
    labelsDf = pandas.DataFrame(1000 * numpy.ones(tsneDf.shape[0]))
    labelsDf.index = tsneDf.index
    labelsDf.columns = ["label"]

    for i in range(len(states)):
        tagged = states[i]
        labelsDf.loc[tagged,"label"] = i
    state_labels = numpy.array(labelsDf.iloc[:,0])
    return state_labels


def plot_states(statesDf, tsneDf, numCols=None, numRows=None, saveFile=None,
                size=10, aspect=1, scale=2):
    if numRows is None:
        if numCols is None:
            numRows = int(round(numpy.sqrt(statesDf.shape[0])))
            rat = numpy.floor(statesDf.shape[0] / float(numRows))
            rem = statesDf.shape[0] - numRows * rat
            numCols = int(rat + rem)
        elif numCols is not None:
            numRows = int(numpy.ceil(float(statesDf.shape[0]) / numCols))

    fig = plt.figure(figsize=(scale * numRows, scale * numCols))

    for ix in range(statesDf.shape[0]):
        ax = fig.add_subplot(numRows, numCols, ix + 1)

        # overlay single state onto tSNE plot
        stateIndex = ix

        group = pandas.DataFrame(numpy.zeros(statesDf.shape[1]))
        group.index = statesDf.columns
        group.columns = ["status"]
        group.loc[statesDf.columns,"status"] = list(statesDf.iloc[stateIndex,:])
        group = numpy.array(group.iloc[:,0])
        ax.set_aspect(aspect)
        ax.scatter(tsneDf.iloc[:,0], tsneDf.iloc[:,1],
                   cmap="bwr", c=group, vmin=-1, vmax=1, s=size)

    if saveFile is not None:
        plt.savefig(saveFile, bbox_inches="tight")


def infer_subtypes(referenceMatrix, primaryMatrix, secondaryMatrix, primaryDictionary,
                   secondaryDictionary,minClusterSize=5,restricted_index=None,
                   tertiaryMatrix=None,tertiaryDictionary=None):

    t1 = time.time()

    logging.info('Beginning subtype inference')
    if restricted_index is not None:
        referenceMatrix = referenceMatrix.loc[restricted_index,:]
        primaryMatrix = primaryMatrix.loc[restricted_index,:]
        secondaryMatrix = secondaryMatrix.loc[restricted_index,:]

    # perform initial subtype clustering
    similarityClusters = f1_decomposition(primaryDictionary,thresholdSFM=0.1)
    similarityClusters = [list(set(cluster)&set(referenceMatrix.columns)) for cluster in similarityClusters]
    initialClasses = [i for i in similarityClusters if len(i)>4]
    if len(initialClasses) == 0:
        logging.info('No subtypes were detected')

    # expand initial subtype clusters
    logginh.info('expand initial subtype clusters')
    centroidClusters, centroidMatrix = centroid_expansion(initialClasses,primaryMatrix,f1Threshold = 0.1,returnCentroids=True) #0.3

    subcentroidClusters = []
    for c in range(len(centroidClusters)):
        tmp_cluster = centroidClusters[c]
        if len(tmp_cluster) < 2*minClusterSize:
            subcentroidClusters.append(tmp_cluster)
            continue

        sampleDictionary = {key:list(set(tmp_cluster)&set(secondaryDictionary[key])) for key in secondaryDictionary}
        sampleMatrix = secondaryMatrix.loc[:,tmp_cluster]

        # perform initial subtype clustering
        similarityClusters = f1_decomposition(sampleDictionary,thresholdSFM=0.1)
        initialClasses = [i for i in similarityClusters if len(i)>4]
        if len(initialClasses)==0:
            subcentroidClusters.append(tmp_cluster)
            continue

        # expand initial subtype clusters
        tmp_centroidClusters, tmp_centroidMatrix = centroid_expansion(initialClasses,sampleMatrix,f1Threshold = 0.1,returnCentroids=True) #0.3

        if len(tmp_centroidClusters) <= 1:
            subcentroidClusters.append(tmp_cluster)
            continue

        for cc in range(len(tmp_centroidClusters)-1):
            new_cluster = tmp_centroidClusters[cc]
            if len(new_cluster) < minClusterSize:
                other_clusters = [tmp_centroidClusters[k] for k in range(len(tmp_centroidClusters)-1) if k!=cc]
                tmp_labels = numpy.array([k for k in range(len(tmp_centroidClusters)-1) if k!=cc])
                new_centroids = get_centroids(other_clusters,referenceMatrix)
                for sample in new_cluster:
                    pearson = pearson_array(numpy.array(new_centroids).T,numpy.array(referenceMatrix.loc[:,sample]))
                    top_hit = numpy.argsort(pearson)[-1]
                    top_label = tmp_labels[top_hit]
                    tmp_centroidClusters[top_label].append(sample)
                tmp_centroidClusters = [tmp_centroidClusters[ii] for ii in range(len(tmp_centroidClusters)) if ii != cc]

            elif len(new_cluster) >= minClusterSize:
                continue

        for ccc in range(len(tmp_centroidClusters)-1):
            subcentroidClusters.append(tmp_centroidClusters[ccc])

        if c == len(centroidClusters)-1:
            subcentroidClusters.append(tmp_centroidClusters[-1])

    if tertiaryMatrix is not None:

        centroidClusters2 = subcentroidClusters
        subcentroidClusters2 = []

        for c in range(len(centroidClusters2)):
            tmp_cluster = centroidClusters2[c]
            if len(tmp_cluster) < 2*minClusterSize:
                subcentroidClusters2.append(tmp_cluster)
                continue
            secondaryDictionary = tertiaryDictionary
            secondaryMatrix = tertiaryMatrix

            sampleDictionary = {key:list(set(tmp_cluster)&set(secondaryDictionary[key])) for key in secondaryDictionary}
            sampleMatrix = secondaryMatrix.loc[:,tmp_cluster]

            # perform initial subtype clustering
            similarityClusters = f1_decomposition(sampleDictionary,thresholdSFM=0.1)
            initialClasses = [i for i in similarityClusters if len(i)>4]
            if len(initialClasses)==0:
                subcentroidClusters2.append(tmp_cluster)
                continue

            # expand initial subtype clusters
            tmp_centroidClusters, tmp_centroidMatrix = centroid_expansion(initialClasses,sampleMatrix,f1Threshold = 0.1,returnCentroids=True) #0.3

            if len(tmp_centroidClusters) <= 1:
                subcentroidClusters2.append(tmp_cluster)
                continue

            for cc in range(len(tmp_centroidClusters)-1):
                new_cluster = tmp_centroidClusters[cc]
                if len(new_cluster) < minClusterSize:
                    other_clusters = [tmp_centroidClusters[k] for k in range(len(tmp_centroidClusters)-1) if k!=cc]
                    tmp_labels = numpy.array([k for k in range(len(tmp_centroidClusters)-1) if k!=cc])
                    new_centroids = get_centroids(other_clusters,referenceMatrix)
                    for sample in new_cluster:
                        pearson = pearson_array(numpy.array(new_centroids).T,
                                                numpy.array(referenceMatrix.loc[:,sample]))
                        top_hit = numpy.argsort(pearson)[-1]
                        top_label = tmp_labels[top_hit]
                        tmp_centroidClusters[top_label].append(sample)
                    tmp_centroidClusters = [tmp_centroidClusters[ii] for ii in range(len(tmp_centroidClusters)) if ii != cc]

                elif len(new_cluster) >= minClusterSize:
                    continue

            for ccc in range(len(tmp_centroidClusters)-1):
                subcentroidClusters2.append(tmp_centroidClusters[ccc])

            if c == len(centroidClusters2)-1:
                subcentroidClusters2.append(tmp_centroidClusters[-1])

        t2 = time.time()
        print("completed subtype inference in {:.2f} minutes".format((t2-t1)/60.))

        return subcentroidClusters2, subcentroidClusters, centroidClusters

    t2 = time.time()
    logging.info("completed subtype inference in {:.2f} minutes".format((t2-t1)/60.))

    return subcentroidClusters, centroidClusters



import numpy
import pandas
from scipy import stats


def __assign_membership(geneset, background, p=0.05):

    cluster = background.loc[geneset, :]
    classNeg1 = len(geneset) - numpy.count_nonzero(cluster + 1, axis=0)
    class0 = len(geneset) - numpy.count_nonzero(cluster, axis=0)
    class1 = len(geneset) - numpy.count_nonzero(cluster - 1, axis=0)
    observations = list(zip(classNeg1, class0, class1))

    highpass = stats.binom.ppf(1 - p / 3, len(geneset), 1./3)
    classes = []
    for i in range(len(observations)):
        check = numpy.where(numpy.array(observations[i]) >= highpass)[0]
        if len(check) > 1:
            check = numpy.array([numpy.argmax(numpy.array(observations[i]))])
        classes.append(check)

    return classes


def make_membership_dictionary(revisedClusters, background, label=2, p=0.05):

    background_genes = set(background.index)
    if label == "excluded":
        members = {}
        for key in revisedClusters.keys():
            tmp_genes = list(set(revisedClusters[key]) & background_genes)
            if len(tmp_genes) > 1:
                assignments = __assign_membership(tmp_genes, background,p=p)
            else:
                assignments = [numpy.array([]) for i in range(background.shape[1])]
            nonMembers = numpy.array([i for i in range(len(assignments)) if len(assignments[i])==0])
            if len(nonMembers) == 0:
                members[key] = []
                continue
            members[key] = list(background.columns[nonMembers])
        return members

    if label == "included":
        members = {}
        for key in revisedClusters.keys():
            tmp_genes = list(set(revisedClusters[key])&background_genes)
            if len(tmp_genes) > 1:
                assignments = __assign_membership(tmp_genes, background, p=p)
            else:
                assignments = [numpy.array([]) for i in range(background.shape[1])]
            included = numpy.array([i for i in range(len(assignments)) if len(assignments[i])!=0])
            if len(included) == 0:
                members[key] = []
                continue

            members[key] = list(background.columns[included])
        return members

    members = {}
    for key in revisedClusters.keys():
        tmp_genes = list(set(revisedClusters[key])&background_genes)
        if len(tmp_genes) > 1:
            assignments = __assign_membership(tmp_genes, background, p=p)
        else:
            assignments = [numpy.array([]) for i in range(background.shape[1])]
        overExpMembers = numpy.array([i for i in range(len(assignments)) if label in assignments[i]])
        if len(overExpMembers) ==0:
            members[key] = []
            continue
        members[key] = list(background.columns[overExpMembers])
    return members


def membership_to_incidence(membershipDictionary, expressionData):

    incidence = numpy.zeros((len(membershipDictionary), expressionData.shape[1]))
    incidence = pandas.DataFrame(incidence)
    incidence.index = membershipDictionary.keys()
    incidence.columns = expressionData.columns
    for key in membershipDictionary.keys():
        samples = membershipDictionary[key]
        incidence.loc[key,samples] = 1

    try:
        orderIndex = numpy.array(incidence.index).astype(int)
        orderIndex = numpy.sort(orderIndex)
    except:
        orderIndex = incidence.index
    try:
        incidence = incidence.loc[orderIndex, :]
    except:
        incidence = incidence.loc[orderIndex.astype(str), :]

    return incidence



import numpy,datetime,pandas,sys,datetime,os
from pkg_resources import Requirement, resource_filename
from collections import Counter
from scipy import stats
from scipy.stats import rankdata

import logging
import traceback as tb

def correct_batch_effects(df, do_preprocess_tpm):
    zscored_expression = zscore(df)
    means = []
    stds = []
    for i in range(zscored_expression.shape[1]):
        mean = numpy.mean(zscored_expression.iloc[:,i])
        std = numpy.std(zscored_expression.iloc[:,i])
        means.append(mean)
        stds.append(std)
    if do_preprocess_tpm and numpy.std(means) >= 0.15:
        zscored_expression = preprocess_tpm(df)
    return zscored_expression


def identifier_conversion(exp_data, conv_table_path=None):
    idMap = pandas.read_table(conv_table_path)
    genetypes = list(set(idMap.iloc[:,2]))
    previousIndex = numpy.array(exp_data.index).astype(str)
    previousColumns = numpy.array(exp_data.columns).astype(str)
    bestMatch = []

    for geneType in genetypes:
        subset = idMap[idMap.iloc[:,2]==geneType]
        subset.index = subset.iloc[:,1]
        mappedGenes = sorted(set(previousIndex)&set(subset.index))
        mappedSamples = sorted(set(previousColumns)&set(subset.index))
        if len(mappedGenes) >= max(10, 0.01 * exp_data.shape[0]):
            if len(mappedGenes)>len(bestMatch):
                bestMatch = mappedGenes
                state = "original"
                gtype = geneType
                continue
        if len(mappedSamples)>=max(10, 0.01 * exp_data.shape[1]):
            if len(mappedSamples)>len(bestMatch):
                bestMatch = mappedSamples
                state = "transpose"
                gtype = geneType
                continue

    mappedGenes = bestMatch
    subset = idMap[idMap.iloc[:,2]==gtype]
    subset.index = subset.iloc[:,1]

    if len(bestMatch) == 0:
        print("Error: Gene identifiers not recognized")

    if state == "transpose":
        exp_data = exp_data.T

    try:
        convertedData = exp_data.loc[mappedGenes,:]
    except:
        tb.print_exc()  # don't silence exceptions !!
        convertedData = exp_data.loc[numpy.array(mappedGenes).astype(int),:]

    conversionTable = subset.loc[mappedGenes,:]
    conversionTable.index = conversionTable.iloc[:,0]
    conversionTable = conversionTable.iloc[:,1]
    conversionTable.columns = ["Name"]

    newIndex = list(subset.loc[mappedGenes,"Preferred_Name"])
    convertedData.index = newIndex

    duplicates = [item for item, count in Counter(newIndex).items() if count > 1]
    singles = list(set(convertedData.index)-set(duplicates))

    corrections = []
    for duplicate in duplicates:
        dupData = convertedData.loc[duplicate,:]
        firstChoice = pandas.DataFrame(dupData.iloc[0,:]).T
        corrections.append(firstChoice)

    if len(corrections)  == 0:
        #print("completed identifier conversion.\n"+str(convertedData.shape[0])+" genes were converted." )
        return convertedData, conversionTable

    correctionsDf = pandas.concat(corrections,axis=0)
    uncorrectedData = convertedData.loc[singles,:]
    convertedData = pandas.concat([uncorrectedData,correctionsDf],axis=0)

    #print("completed identifier conversion.\n"+str(convertedData.shape[0])+" genes were converted." )

    return convertedData, conversionTable


"""
def identifier_conversion(expression_data, conversion_table_path=None):

    # if not specified, read conversion table from package data
    if conversion_table_path is None:
        conversion_table_path = resource_filename(Requirement.parse("miner2"),'miner2/data/identifier_mappings.txt')

    id_map = pandas.read_csv(conversion_table_path,sep='\t')

    gene_types = list(set(id_map.iloc[:,2]))

    ### ALO.2019.06.19. Addressing dot transcripts in ENSEMBL annotations from GDC portal
    #previous_index = numpy.array(expression_data.index).astype(str) # this line is obsolete. Kept for reference. To be removed in the future
    original_IDs=numpy.array(expression_data.index).astype(str)
    logging.info('evaluate removing transcript isoform information')
    original_IDs_list=list(original_IDs)
    original_annotation=numpy.array([element.split('.')[0] for element in original_IDs_list])
    if original_IDs.shape == original_annotation.shape:
        logging.info('no information loss')
        substitution={element:element.split('.')[0] for element in original_IDs_list}
        expression_data.rename(index=substitution,inplace=True)
    else:
        logging.warn('some gene IDs have multiple transcript isoforms. Initial {} versus new {} dimensions'.format(original_IDs.shape,original_annotation.shape))
    ### end ALO

    previous_columns = numpy.array(expression_data.columns).astype(str)
    best_match = []

    # WW: This weird loop actually tries to determine whether the expression
    # matrix is original (in which case the genes are the rows) or a transpose
    # (in which case the genes are the columns). It does does so by trying to
    # map the genes in the id map to both the rows and columns and if there
    # are multiple matches in either it will determine those to be the genes.
    # It also tries to find the gene type with the most gene matches and build
    # the conversion table based on the gene type with the most gene matches
    #
    # The code seems to have multiple issues, and is very sensitive to changes
    # we should see if we can simplify it

    for gene_type in gene_types:
        subset = id_map[id_map.iloc[:,2] == gene_type]
        subset.index = subset.iloc[:,1]
        mapped_genes = list(set(original_annotation) & set(subset.index))
        mapped_samples = list(set(previous_columns) & set(subset.index))

        if len(mapped_genes) >= max(10, 0.01 * expression_data.shape[0]):
            if len(mapped_genes) > len(best_match):
                best_match = mapped_genes
                state = "original"
                gtype = gene_type
                logging.info('expression data arrangement detected: genes as rows ({}) and samples as columns ({})'.format(expression_data.shape[0],expression_data.shape[1]))
                continue

        if len(mapped_samples) >= max(10, 0.01 * expression_data.shape[1]):
            if len(mapped_samples) > len(best_match):
                best_match = mapped_samples
                state = "transpose"
                gtype = gene_type
                logging.info('expression data arrangement detected: genes as columns ({}) and samples as rows ({})'.format(expression_data.shape[0],expression_data.shape[1]))
                continue

    if len(best_match) == 0:
        raise Exception("Error: Gene identifiers not recognized")

    # ALO/WW: Without sorting the mapped_genes, the values will differ in
    # multiple places, we should check why this has such a great impact
    mapped_genes = sorted(best_match)

    subset = id_map[id_map.iloc[:,2]==gtype]
    subset.index = subset.iloc[:,1]

    if state == "transpose":
        logging.info('transpose expression data')
        expression_data = expression_data.T

    converted_data = expression_data.loc[mapped_genes,:]

    conversion_table = subset.loc[mapped_genes,:]
    conversion_table.index = conversion_table.iloc[:,0]
    conversion_table = conversion_table.iloc[:,1]
    conversion_table.columns = ["Name"]

    new_index = list(subset.loc[mapped_genes, "Preferred_Name"])
    converted_data.index = new_index

    # WW: Due to ambiguous mapping converted_data will potentially have
    # multiple rows containing the same row name, so this conflict
    # needs to be resolved
    duplicates = [item for item, count in Counter(new_index).items() if count > 1]
    singles = list(set(converted_data.index) - set(duplicates))

    # WW: with the sorting of the duplicate and singles lists we are eliminating the
    # order ambiguity that exists when we switch between Python 2 and 3, but
    # in actuality the fundamental problem is the gene mapping
    duplicates.sort()
    singles.sort()

    corrections = []

    if len(duplicates) > 0:
        logging.warn("There were %d ambiguously mapped genes !", len(duplicates))

    # WW: The way duplicates are resolved is by retrieving the duplicate rows
    # as individual DataFrames and picking the first one
    # TODO: we might have to rethink the strategy because the ambiguous row names
    # were introduced by the way the rows were remapped and we actually need to
    # map them to individual genes
    for duplicate in duplicates:
        dup_data = converted_data.loc[duplicate,:]
        first_choice = pandas.DataFrame(dup_data.iloc[0,:]).T
        corrections.append(first_choice)

    if len(corrections) > 0:
        corrections_df = pandas.concat(corrections, axis=0)
        uncorrected_data = converted_data.loc[singles,:]
        converted_data = pandas.concat([uncorrected_data, corrections_df], axis=0)

    logging.info("{} out of {} gene names converted to ENSEMBL IDs".format(converted_data.shape[0], expression_data.shape[0]))

    return converted_data,conversion_table
"""

def entropy(vector):

    data = numpy.array(vector)
    hist = numpy.histogram(data, bins=50)[0]
    length = len(hist)

    if length <= 1:
        return 0

    counts = numpy.bincount(hist)
    probs = [float(i) / length for i in counts]
    n_classes = numpy.count_nonzero(probs)

    if n_classes <= 1:
        return 0

    ent = 0.

    # Compute standard entropy.
    for i in probs:
        if i > 0:
            ent -= float(i) * numpy.log(i)
    return ent

def main(input_path,conversion_table_path=None, do_preprocess_tpm=True):

    # first detect if it's a dataframe or dir or a file to parse data differently
    if isinstance(input, pandas.DataFrame):
        raw_expression = input
    elif os.path.isfile(input_path) == True:
        logging.info('detected expression data file')
        raw_expression = read_file_to_df(input_path)
    elif os.path.isdir(input_path) == True:
        logging.info('detected folder from GDC')
        raw_expression_data = read_expression_from_GDC_download(input_path)
        raw_expression = transform_to_FPKM(raw_expression_data,fpkm_threshold=1,min_fraction_above_threshold=0.5,highly_expressed=False)
    else:
        raise Exception("Error: Unable to identify if input path is directory or expression file.")
    logging.info("expression data recovered: {} features by {} samples".format(raw_expression.shape[0], raw_expression.shape[1]))

    # data transformations
    logging.info("expression data transformation")
    raw_expression_zero_filtered = remove_null_rows(raw_expression)
    zscored_expression = correct_batch_effects(raw_expression_zero_filtered, do_preprocess_tpm)

    ("gene ID conversion")
    expression_data, conversion_table = identifier_conversion(zscored_expression,conversion_table_path)
    logging.info("working expression data: {} features by {} samples".format(expression_data.shape[0], expression_data.shape[1]))

    return expression_data, conversion_table

def read_expression_from_GDC_download(directory):

    sample_dfs=[]
    for dirName, subdirList, fileList in os.walk(directory):
        for fname in fileList:
            if fname[0] != '.':
                extension=fname.split(".")[-1]
                if extension == 'gz':
                    path=os.path.join(directory,dirName,fname)
                    df=pandas.read_csv(path,compression='gzip',index_col=0,header=None,sep='\t',quotechar='"')
                    df.columns=[fname.split(".")[0]]
                    sample_dfs.append(df)

    expressionData = pandas.concat(sample_dfs,axis=1)

    return expressionData


def read_file_to_df(filename):
    """read expression file into a Pandas DataFrame depending on the file extension"""
    extension = filename.split(".")[-1]
    if extension == "csv":
        df = pandas.read_csv(filename, index_col=0, header=0)
        num_rows, num_cols = df.shape
        if num_cols == 0:
            df = pandas.read_csv(filename, index_col=0, header=0, sep="\t")
    elif extension == "txt":
        df = pandas.read_csv(filename, index_col=0, header=0, sep="\t")
        num_rows, num_cols = df.shape
        if num_cols == 0:
            df = pandas.read_csv(filename, index_col=0, header=0)
    return df


def remove_null_rows(df):
    minimum = numpy.percentile(df, 0)
    if minimum == 0:
        filtered_df = df.loc[df.sum(axis=1)>0,:]
    else:
        filtered_df = df
    return filtered_df


def preprocess_tpm(tpm):
    cutoff = stats.norm.ppf(0.00001)
    tmp_array_raw = numpy.array(tpm)
    keep = []
    keepappend = keep.append
    for i in range(0, tmp_array_raw.shape[0]):
        if numpy.count_nonzero(tmp_array_raw[i,:]) >= round(float(tpm.shape[1])*0.5):
            keepappend(i)

    tpm_zero_filtered = tmp_array_raw[keep,:]
    tpm_array = numpy.array(tpm_zero_filtered)
    positive_medians = []

    for i in range(0, tpm_array.shape[1]):
        tmp1 = tpm_array[:,i][tpm_array[:,i]>0]
        positive_medians.append(numpy.median(tmp1))

    # 2^10 - 1 = 1023
    scale_factors = [float(1023)/positive_medians[i] for i in range(0,len(positive_medians))]

    tpm_scale = numpy.zeros(tpm_array.shape)
    for i in range(0, tpm_scale.shape[1]):
        tpm_scale[:,i] = tpm_array[:,i]*scale_factors[i]

    tpm_scale_log2 = numpy.zeros(tpm_scale.shape)
    for i in range(0, tpm_scale_log2.shape[1]):
        tpm_scale_log2[:,i] = numpy.log2(tpm_scale[:,i]+1)

    tpm_filtered_df = pandas.DataFrame(tpm_scale_log2)
    tpm_filtered_df.columns = list(tpm.columns)
    tpm_filtered_df.index = list(numpy.array(tpm.index)[keep])

    qn_tpm_filtered = quantile_norm(tpm_filtered_df,axis=0)
    qn_tpm = quantile_norm(qn_tpm_filtered,axis=1)

    qn_tpm_array = numpy.array(qn_tpm)

    tpm_z = numpy.zeros(qn_tpm_array.shape)
    for i in range(0, tpm_z.shape[0]):
        tmp = qn_tpm_array[i,:][qn_tpm_array[i,:]>0]
        mean = numpy.mean(tmp)
        std = numpy.std(tmp)
        for j in range(0,tpm_z.shape[1]):
            tpm_z[i,j] = float(qn_tpm_array[i,j] - mean)/std
            if tpm_z[i,j] < -4:
                tpm_z[i,j] = cutoff

    tpm_entropy = []
    for i in range(0,tpm_z.shape[0]):
        tmp = entropy(tpm_z[i,:])
        tpm_entropy.append(tmp)

    tpmz_df = pandas.DataFrame(tpm_z)
    tpmz_df.columns = list(tpm.columns)
    tpmz_df.index = list(numpy.array(tpm.index)[keep])


    ent = pandas.DataFrame(tpm_entropy)
    ent.index = list(tpmz_df.index)
    ent.columns = ['entropy']

    tpm_ent_df = pandas.concat([tpmz_df,ent],axis=1)

    tpm_entropy_sorted = tpm_ent_df.sort_values(by='entropy',ascending=False)

    tmp = tpm_entropy_sorted[tpm_entropy_sorted.loc[:,'entropy']>=0]
    tpm_select = tmp.iloc[:,0:-1]

    return tpm_select


def quantile_norm(df,axis=1):

    if axis == 1:
        array = numpy.array(df)

        ranked_array = numpy.zeros(array.shape)
        for i in range(0,array.shape[0]):
            ranked_array[i,:] = rankdata(array[i,:],method='min') - 1

        sorted_array = numpy.zeros(array.shape)
        for i in range(0,array.shape[0]):
            sorted_array[i,:] = numpy.sort(array[i,:])

        qn_values = numpy.nanmedian(sorted_array,axis=0)

        quant_norm_array = numpy.zeros(array.shape)
        for i in range(0,array.shape[0]):
            for j in range(0,array.shape[1]):
                quant_norm_array[i,j] = qn_values[int(ranked_array[i,j])]

        quant_norm = pandas.DataFrame(quant_norm_array)
        quant_norm.columns = list(df.columns)
        quant_norm.index = list(df.index)

    if axis == 0:
        array = numpy.array(df)

        ranked_array = numpy.zeros(array.shape)
        for i in range(0,array.shape[1]):
            ranked_array[:,i] = rankdata(array[:,i],method='min') - 1

        sorted_array = numpy.zeros(array.shape)
        for i in range(0,array.shape[1]):
            sorted_array[:,i] = numpy.sort(array[:,i])

        qn_values = numpy.nanmedian(sorted_array,axis=1)

        quant_norm_array = numpy.zeros(array.shape)
        for i in range(0,array.shape[0]):
            for j in range(0,array.shape[1]):
                quant_norm_array[i,j] = qn_values[int(ranked_array[i,j])]

        quant_norm = pandas.DataFrame(quant_norm_array)
        quant_norm.columns = list(df.columns)
        quant_norm.index = list(df.index)

    return quant_norm

def transform_to_FPKM(expression_data,fpkm_threshold=1,min_fraction_above_threshold=0.5,highly_expressed=False,quantile_normalize=False):

    median = numpy.median(numpy.median(expression_data,axis=1))
    expDataCopy = expression_data.copy()
    expDataCopy[expDataCopy<fpkm_threshold]=0
    expDataCopy[expDataCopy>0]=1
    cnz = numpy.count_nonzero(expDataCopy,axis=1)
    keepers = numpy.where(cnz>=int(min_fraction_above_threshold*expDataCopy.shape[1]))[0]
    threshold_genes = expression_data.index[keepers]
    expDataFiltered = expression_data.loc[threshold_genes,:]

    if highly_expressed is True:
        median = numpy.median(numpy.median(expDataFiltered,axis=1))
        expDataCopy = expDataFiltered.copy()
        expDataCopy[expDataCopy<median]=0
        expDataCopy[expDataCopy>0]=1
        cnz = numpy.count_nonzero(expDataCopy,axis=1)
        keepers = numpy.where(cnz>=int(0.5*expDataCopy.shape[1]))[0]
        median_filtered_genes = expDataFiltered.index[keepers]
        expDataFiltered = expression_data.loc[median_filtered_genes,:]

    if quantile_normalize is True:
        expDataFiltered = quantile_norm(expDataFiltered,axis=0)

    finalExpData = pandas.DataFrame(numpy.log2(expDataFiltered+1))
    finalExpData.index = expDataFiltered.index
    finalExpData.columns = expDataFiltered.columns

    return finalExpData

def zscore(expressionData):
    zero = numpy.percentile(expressionData,0)
    meanCheck = numpy.mean(expressionData[expressionData>zero].mean(axis=1,skipna=True))
    if meanCheck<0.1:
        return expressionData
    means = expressionData.mean(axis=1,skipna=True)
    stds = expressionData.std(axis=1,skipna=True)
    try:
        transform = ((expressionData.T - means)/stds).T
    except:
        tb.print_exc()  # don't silence exceptions !!
        passIndex = numpy.where(stds>0)[0]
        transform = ((expressionData.iloc[passIndex,:].T - means[passIndex])/stds[passIndex]).T
    return transform


def background_df(expression_data):

    low = numpy.percentile(expression_data, 100. / 3, axis=0)
    high = numpy.percentile(expression_data, 200. / 3, axis=0)
    even_cuts = list(zip(low, high))

    bkgd = expression_data.copy()
    for i in range(bkgd.shape[1]):
        low_cut = even_cuts[i][0]
        high_cut = even_cuts[i][1]
        bkgd.iloc[:, i][bkgd.iloc[:, i] >= high_cut] = 1
        bkgd.iloc[:, i][bkgd.iloc[:, i] <= low_cut] = -1
        bkgd.iloc[:, i][numpy.abs(bkgd.iloc[:, i]) != 1] = 0

    return bkgd



import os
import pandas as pd
import numpy as np
import warnings

import sklearn
from sklearn.model_selection import train_test_split
from sklearn import metrics
import matplotlib.pyplot as plt
import json
import logging

from miner2 import survival, preprocess, biclusters, miner


def generate_prediction_matrix(srv, mtrx, high_risk_cutoff=0.2):

    srv = srv.copy()
    srv.sort_values(by='GuanScore',ascending=False,inplace=True)

    highRiskSamples = list(srv.index[0:int(high_risk_cutoff*srv.shape[0])])
    lowRiskSamples = list(srv.index[int(high_risk_cutoff*srv.shape[0]):])

    hrFlag = pd.DataFrame(np.ones((len(highRiskSamples),1)).astype(int))
    hrFlag.index = highRiskSamples
    hrFlag.columns = ["HR_FLAG"]

    lrFlag = pd.DataFrame(np.zeros((len(lowRiskSamples),1)).astype(int))
    lrFlag.index = lowRiskSamples
    lrFlag.columns = ["HR_FLAG"]

    # This crashes when none of the samples in highRiskSamples are
    # found in the mtrx columns, so we fall back to return None
    try:
        hrMatrix = pd.concat([mtrx.loc[:,highRiskSamples].T,hrFlag],axis=1)
        hrMatrix.columns = np.array(hrMatrix.columns).astype(str)
    except:
        hrMatrix = None

    # This crashes when none of the samples in lowRiskSamples are
    # found in the mtrx columns, so we fall back to return None
    try:
        lrMatrix = pd.concat([mtrx.loc[:,lowRiskSamples].T,lrFlag],axis=1)
        lrMatrix.columns = np.array(lrMatrix.columns).astype(str)
    except:
        lrMatrix = None

    return hrMatrix, lrMatrix


def prediction_matrix(membership_datasets,survival_datasets,high_risk_cutoff=0.20):
    hr_matrices = []
    lr_matrices = []

    for i in range(len(membership_datasets)):
        hrmatrix, lrmatrix = generate_prediction_matrix(survival_datasets[i],
                                                        membership_datasets[i],
                                                        high_risk_cutoff=high_risk_cutoff)
        if hrmatrix is not None:
            hr_matrices.append(hrmatrix)
        if lrmatrix is not None:
            lr_matrices.append(lrmatrix)

    hrMatrixCombined = pd.concat(hr_matrices,axis=0)
    lrMatrixCombined = pd.concat(lr_matrices,axis=0)
    predictionMat = pd.concat([hrMatrixCombined,lrMatrixCombined],axis=0)
    return predictionMat


def risk_stratification(lbls, mtrx, guan_srv, survival_tag, classifier,
                        resultsDirectory=None, plot_all=False, guan_rank=False,
                        high_risk_cutoffs=None, plot_any=True):
    warnings.filterwarnings("ignore")

    guan_srv = guan_srv.loc[list(set(guan_srv.index)&set(mtrx.columns)),:]
    if plot_any is True:
        f, (ax1, ax2) = plt.subplots(1, 2, sharey=False)
        f.tight_layout(pad=1.08)
        f.set_figwidth(10)
        f.set_figheight(4)

    predicted_probabilities = classifier.predict_proba(np.array(mtrx.T))[:,1]
    predicted_probabilities_df = pd.DataFrame(predicted_probabilities)
    predicted_probabilities_df.index = mtrx.columns
    predicted_probabilities_df.columns = ["probability_high_risk"]

    srv = guan_srv.iloc[:,0:2]
    srv_observed = guan_srv[guan_srv.iloc[:,1]==1]
    srv_unobserved = guan_srv[guan_srv.iloc[:,1]==0]

    if high_risk_cutoffs is None:
        high_risk_cutoffs = np.percentile(list(srv_observed.iloc[:,0]),[10,15,20,25,30])

    aucs = []
    cutoffs = []
    tpr_list = []
    fpr_list = []
    prec = []
    rec = []
    for i in range(len(high_risk_cutoffs)):#range(0,max(list(srv.iloc[:,0]))+interval,interval):
        if guan_rank is True:
            percentile = 10+i*(20.0/(len(high_risk_cutoffs)-1))
            number_samples = int(np.ceil(guan_srv.shape[0]*(percentile/100.0)))
            cutoff = guan_srv.iloc[number_samples,0]
            true_hr = guan_srv.index[0:number_samples]
            true_lr = guan_srv.index[number_samples:]
            srv_total = guan_srv.copy()

        elif guan_rank is not True:
            cutoff = high_risk_cutoffs[i]
            srv_extended = srv_unobserved[srv_unobserved.iloc[:,0]>=cutoff]
            if srv_extended.shape[0] > 0:
                srv_total = pd.concat([srv_observed,srv_extended],axis=0)
            elif srv_extended.shape[0] == 0:
                srv_total = srv_observed.copy()
            pass_index = np.where(srv_total.iloc[:,0]<cutoff)[0]
            true_hr = []
            if len(pass_index) > 0:
                true_hr = srv_total.index[pass_index]
            true_lr = list(set(srv_total.index)-set(true_hr))


        # use predicted_probabilities_df against true_hr, true_lr to
        # compute precision and recall from sklearn.metrics
        tpr = []
        fpr = []
        precisions = []
        recalls = []
        for threshold in np.arange(0,1.02,0.01):
            model_pp = predicted_probabilities_df.copy()
            model_pp = model_pp.loc[list(set(model_pp.index)&set(srv_total.index)),:]
            predicted_hr = model_pp.index[np.where(model_pp.iloc[:,0]>=threshold)[0]]
            predicted_lr = list(set(model_pp.index)-set(predicted_hr))

            tp = set(true_hr)&set(predicted_hr)
            fp = set(true_lr)&set(predicted_hr)
            allpos = set(true_hr)
            tn = set(true_lr)&set(predicted_lr)
            fn = set(true_hr)&set(predicted_lr)
            allneg = set(true_lr)

            if len(allpos) == 0:
                tp_rate = 0
                precision = 0
                recall=0
            elif len(allpos) > 0:
                tp_rate = len(tp)/float(len(allpos))

                if len(tp) + len(fp) > 0:
                    precision = len(tp)/float(len(tp) + len(fp))
                elif len(tp) + len(fp) == 0:
                    precision = 0

                if len(tp) +len(fn) > 0:
                    recall = len(tp)/float(len(tp) +len(fn))
                elif len(tp) +len(fn) == 0:
                    recall = 0
            if len(allneg) == 0:
                tn_rate = 0
            elif len(allneg) > 0:
                tn_rate = len(tn)/float(len(allneg))

            tpr.append(tp_rate)
            fpr.append(1-tn_rate)

            precisions.append(precision)
            recalls.append(recall)

        if plot_all:
            plt.figure()
            plt.plot(fpr,tpr)
            plt.plot(np.arange(0,1.01,0.01),np.arange(0,1.01,0.01),"--")
            plt.ylim(-0.05,1.05)
            plt.xlim(-0.05,1.05)
            plt.title('ROC curve, cutoff = {:d}'.format(int(cutoff)))

        area = metrics.auc(fpr,tpr)
        aucs.append(area)
        cutoffs.append(cutoff)
        tpr_list.append(tpr)
        fpr_list.append(fpr)
        prec.append(precisions)
        rec.append(recalls)

    integrated_auc = np.mean(aucs)

    #print('classifier has integrated AUC of {:.3f}'.format(integrated_auc))

    tpr_stds = np.std(np.vstack(tpr_list),axis=0)
    tpr_means = np.mean(np.vstack(tpr_list),axis=0)
    fpr_means = np.mean(np.vstack(fpr_list),axis=0)

    if plot_any:
        ax1.fill_between(fpr_means, tpr_means-tpr_stds, tpr_means+tpr_stds,color=[0,0.4,0.6],alpha=0.3)
        ax1.plot(fpr_means,tpr_means,color=[0,0.4,0.6],LineWidth=1.5)
        ax1.plot(np.arange(0,1.01,0.01),np.arange(0,1.01,0.01),"--",color=[0.2,0.2,0.2])
        ax1.set_ylim(-0.05,1.05)
        ax1.set_xlim(-0.05,1.05)
        ax1.set_title('Integrated AUC = {:.2f}'.format(integrated_auc))
        ax1.set_ylabel('Sensitivity',FontSize=14)
        ax1.set_xlabel('1-Specificity',FontSize=14)

    hr_dt = mtrx.columns[lbls.astype(bool)]
    lr_dt = mtrx.columns[(1-lbls).astype(bool)]

    kmTag = "decision_tree"
    kmFilename = ("_").join([survival_tag,kmTag,"high-risk",".pdf"])

    groups = [hr_dt,lr_dt]
    labels = ["High-risk","Low-risk"]

    cox_vectors = []
    srv_set = set(srv.index)
    for i in range(len(groups)):
        group = groups[i]
        patients = list(set(group)&srv_set)
        tmp_df = pd.DataFrame(np.zeros(srv.shape[0]))
        tmp_df.index = srv.index
        tmp_df.columns = [labels[i]]
        tmp_df.loc[patients,labels[i]] = 1
        cox_vectors.append(tmp_df)

    pre_cox = pd.concat(cox_vectors,axis=1).T
    pre_cox.head(5)

    cox_dict = survival.parallel_member_survival_analysis(membershipDf=pre_cox,
                                                          numCores=1,
                                                          survivalPath="",
                                                          survivalData=srv)
    #print('Risk stratification of '+survival_tag+' has Hazard Ratio of {:.2f}'.format(cox_dict['High-risk'][0]))

    hazard_ratio = cox_dict['High-risk'][0]
    if plot_any is True:
        if resultsDirectory is not None:
            plotName = os.path.join(resultsDirectory,kmFilename)
            survival.kmplot(srv=srv,groups=groups,labels=labels,xlim_=(-100,1750),filename=plotName)
            plt.title('Dataset: '+survival_tag+'; HR: {:.2f}'.format(cox_dict['High-risk'][0]))

        elif resultsDirectory is None:
            survival.kmplot(srv=srv,groups=groups,labels=labels,xlim_=(-100,1750),filename=None)
            plt.title('Dataset: '+survival_tag+'; HR: {:.2f}'.format(cox_dict['High-risk'][0]))

    return aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec


def generate_predictor(membership_datasets, survival_datasets, dataset_labels,
                       iterations=20, method='xgboost', n_estimators=100,
                       output_directory=None, best_state=None, test_only=True,
                       separate_results=True, metric='roc_auc', class1_proportion=0.20,
                       test_proportion=0.35, colsample_bytree=1, subsample=1,
                       verbose=False):
    """
    Computes a classifier object from the specified input data sets.
    The result is either an xgboost.XGBClassifier or a sklearn.tree.DecisionTreeClassifier
    """
    if method == 'xgboost':
        # prevents kernel from dying when running XGBClassifier
        os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'
        from xgboost import XGBClassifier

    elif method=='decisionTree':
        from sklearn.tree import DecisionTreeClassifier

    predictionMat = prediction_matrix(membership_datasets,survival_datasets,
                                      high_risk_cutoff=class1_proportion)

    pm_keys = sorted(predictionMat.keys())
    cox_hr = [predictionMat[i][0] for i in pm_keys]
    cox_p = [predictionMat[i][1] for i in pm_keys]
    cox_combined_states_df = pd.DataFrame(np.vstack([cox_hr,cox_p]).T)
    #cox_combined_states_df.index = predictionMat.index
    cox_combined_states_df.columns = ["HR","p-value"]
    cox_combined_states_df.sort_values(by="HR",ascending=False,inplace=True)
    cox_combined_states_df.sort_index(inplace=True)
    cox_combined_states_df.to_csv(os.path.join(output_directory, 'CoxProportionalHazardsRegulons.csv'))

    X = np.array(predictionMat.iloc[:,0:-1])
    Y = np.array(predictionMat.iloc[:,-1])
    X = X.astype('int')
    Y = Y.astype('int')

    samples_ = np.array(predictionMat.index)

    if best_state is None:
        mean_aucs = []
        mean_hrs = []
        pct_labeled = []
        for rs in range(iterations):
            X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size = test_proportion, random_state = rs)
            X_train_columns, X_test_columns, y_train_samples, y_test_samples = train_test_split(X, samples_, test_size = test_proportion, random_state = rs)

            train_datasets = []
            test_datasets = []
            for td in range(len(membership_datasets)):
                dataset = membership_datasets[td]
                train_members = list(set(dataset.columns)&set(y_train_samples))
                test_members = list(set(dataset.columns)&set(y_test_samples))
                train_datasets.append(dataset.loc[:,train_members])
                test_datasets.append(dataset.loc[:,test_members])

            if method=='xgboost':
                eval_set = [(X_train, y_train), (X_test, y_test)]
                clf = XGBClassifier(n_jobs=1,random_state=12,n_estimators=n_estimators,colsample_bytree=colsample_bytree,subsample=subsample)
                clf.fit(X_train, y_train, early_stopping_rounds=10, eval_metric="auc", eval_set=eval_set, verbose=False)
            elif method=='decisionTree':
                clf = DecisionTreeClassifier(criterion = "gini", random_state = 12, max_depth=6, min_samples_leaf=5)
                clf.fit(X_train, y_train)

            train_predictions = []
            test_predictions = []
            for p in range(len(membership_datasets)):
                tmp_train_predictions = clf.predict(np.array(train_datasets[p].T))
                tmp_test_predictions = clf.predict(np.array(test_datasets[p].T))
                train_predictions.append(tmp_train_predictions)
                test_predictions.append(tmp_test_predictions)

            if test_only:
                scores = []
                hrs = []
                for j in range(len(test_datasets)):
                    mtrx = test_datasets[j]
                    guan_srv = survival_datasets[j]
                    survival_tag = dataset_labels[j]
                    lbls = test_predictions[j]
                    aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=False)
                    score = np.mean(aucs)
                    scores.append(score)
                    hrs.append(hazard_ratio)
                    pct_labeled.append(100*sum(lbls)/float(len(lbls)))

                mean_auc = np.mean(scores)
                mean_hr = np.mean(hrs)
                mean_aucs.append(mean_auc)
                mean_hrs.append(mean_hr)
                #print(rs,mean_auc,mean_hr)

            elif test_only is False:
                scores = []
                hrs = []
                for j in range(len(test_datasets)):
                    mtrx = test_datasets[j]
                    guan_srv = survival_datasets[j]
                    survival_tag = dataset_labels[j]
                    lbls = test_predictions[j]
                    aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=False)
                    score = np.mean(aucs)
                    scores.append(score)
                    hrs.append(hazard_ratio)

                    mtrx = train_datasets[j]
                    lbls = train_predictions[j]
                    aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=False)
                    score = np.mean(aucs)
                    scores.append(score)
                    hrs.append(hazard_ratio)

                mean_auc = np.mean(scores)
                mean_hr = np.mean(hrs)
                mean_aucs.append(mean_auc)
                mean_hrs.append(mean_hr)
                #print(rs,mean_auc,mean_hr)

        if metric == 'roc_auc':
            best_state = np.argsort(np.array(mean_aucs))[-1]
        elif metric == 'hazard_ratio':
            best_state = np.argsort(np.array(mean_hrs))[-1]

    X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size = test_proportion, random_state = best_state)
    X_train_columns, X_test_columns, y_train_samples, y_test_samples = train_test_split(X, samples_, test_size = test_proportion, random_state = best_state)

    train_datasets = []
    test_datasets = []
    for td in range(len(membership_datasets)):
        dataset = membership_datasets[td]
        train_members = list(set(dataset.columns)&set(y_train_samples))
        test_members = list(set(dataset.columns)&set(y_test_samples))
        train_datasets.append(dataset.loc[:,train_members])
        test_datasets.append(dataset.loc[:,test_members])

    if method=='xgboost':
        eval_set = [(X_train, y_train), (X_test, y_test)]
        clf = XGBClassifier(n_jobs=1,random_state=12,n_estimators=n_estimators,colsample_bytree=colsample_bytree,subsample=subsample)
        clf.fit(X_train, y_train, early_stopping_rounds=10, eval_metric="auc", eval_set=eval_set, verbose=False)
    elif method=='decisionTree':
        clf = DecisionTreeClassifier(criterion = "gini", random_state = 12, max_depth=6, min_samples_leaf=5)
        clf.fit(X_train, y_train)

    train_predictions = []
    test_predictions = []
    for p in range(len(membership_datasets)):
        tmp_train_predictions = clf.predict(np.array(train_datasets[p].T))
        tmp_test_predictions = clf.predict(np.array(test_datasets[p].T))
        train_predictions.append(tmp_train_predictions)
        test_predictions.append(tmp_test_predictions)

    mean_aucs = []
    mean_hrs = []
    if test_only:
        scores = []
        hrs = []
        pct_labeled = []
        for j in range(len(test_datasets)):
            mtrx = test_datasets[j]
            guan_srv = survival_datasets[j]
            survival_tag = dataset_labels[j]
            lbls = test_predictions[j]
            aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=False)
            score = np.mean(aucs)
            scores.append(score)
            hrs.append(hazard_ratio)
            pct_labeled.append(100*sum(lbls)/float(len(lbls)))

        mean_auc = np.mean(scores)
        mean_hr = np.mean(hrs)
        mean_aucs.append(mean_auc)
        mean_hrs.append(mean_hr)
        precision_matrix = np.vstack(prec)
        recall_matrix = np.vstack(rec)
        #print(best_state,mean_auc,mean_hr)

    else:
        scores = []
        hrs = []
        pct_labeled = []
        for j in range(len(test_datasets)):
            mtrx = test_datasets[j]
            guan_srv = survival_datasets[j]
            survival_tag = dataset_labels[j]
            lbls = test_predictions[j]
            aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=False)
            score = np.mean(aucs)
            scores.append(score)
            hrs.append(hazard_ratio)
            pct_labeled.append(100*sum(lbls)/float(len(lbls)))

            mtrx = train_datasets[j]
            lbls = train_predictions[j]
            aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=False)
            score = np.mean(aucs)
            scores.append(score)
            hrs.append(hazard_ratio)

        mean_auc = np.mean(scores)
        mean_hr = np.mean(hrs)
        mean_aucs.append(mean_auc)
        mean_hrs.append(mean_hr)
        precision_matrix = np.vstack(prec)
        recall_matrix = np.vstack(rec)
        #print(best_state,mean_auc,mean_hr)

    train_predictions = []
    test_predictions = []
    predictions = []
    #add print for percent labeled high-risk
    for p in range(len(membership_datasets)):
        tmp_train_predictions = clf.predict(np.array(train_datasets[p].T))
        tmp_test_predictions = clf.predict(np.array(test_datasets[p].T))
        tmp_predictions = clf.predict(np.array(membership_datasets[p].T))
        train_predictions.append(tmp_train_predictions)
        test_predictions.append(tmp_test_predictions)
        predictions.append(tmp_predictions)

    if not separate_results:
        for j in range(len(membership_datasets)):
            mtrx = membership_datasets[j]
            guan_srv = survival_datasets[j]
            survival_tag = dataset_labels[j]
            lbls = predictions[j]

            percent_classified_hr = 100 * sum(lbls) / float(len(lbls))
            logging.info('classified {:.1f} percent of population as high-risk'.format(percent_classified_hr))

            aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=True)
            if output_directory is not None:
                plt.savefig(os.path.join(output_directory,('_').join([survival_tag,method,metric,'survival_predictions.pdf'])),bbox_inches='tight')

    else:
        for j in range(len(membership_datasets)):
            guan_srv = survival_datasets[j]
            survival_tag = dataset_labels[j]

            mtrx = train_datasets[j]
            lbls = train_predictions[j]

            percent_classified_hr = 100*sum(lbls)/float(len(lbls))
            logging.info('classified {:.1f} percent of training population as high-risk'.format(percent_classified_hr))

            aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=True)
            if output_directory is not None:
                plt.savefig(os.path.join(output_directory,('_').join([survival_tag,method,metric,'training_survival_predictions.pdf'])),bbox_inches='tight')

            mtrx = test_datasets[j]
            lbls = test_predictions[j]

            percent_classified_hr = 100*sum(lbls)/float(len(lbls))
            logging.info('classified {:.1f} percent of test population as high-risk'.format(percent_classified_hr))

            aucs, cutoffs, tpr_list, fpr_list, hazard_ratio, prec, rec = risk_stratification(lbls,mtrx,guan_srv,survival_tag,clf,guan_rank=False,resultsDirectory=None,plot_all=False,plot_any=True)
            if output_directory is not None:
                plt.savefig(os.path.join(output_directory,('_').join([survival_tag,method,metric,'test_survival_predictions.pdf'])),bbox_inches='tight')

    nextIteration = []
    class0 = []
    class1 = []
    for p in range(len(membership_datasets)):
        tmp_predictions = clf.predict(np.array(membership_datasets[p].T))
        tmp_class_0 = membership_datasets[p].columns[(1-np.array(tmp_predictions)).astype(bool)]
        tmp_class_1 = membership_datasets[p].columns[np.array(tmp_predictions).astype(bool)]
        nextIteration.append(membership_datasets[p].loc[:,tmp_class_0])
        class0.append(tmp_class_0)
        class1.append(tmp_class_1)

    #print(best_state)

    if best_state is not None:
        return clf, class0, class1, mean_aucs, mean_hrs, pct_labeled, precision_matrix, recall_matrix

    return clf, class0, class1, mean_aucs, mean_hrs, pct_labeled, precision_matrix, recall_matrix


def load_test_set(exp_path, reference_modules, conv_table_path):
    exp_data = pd.read_csv(exp_path, index_col=0,header=0)
    exp_data, _ = preprocess.identifier_conversion(exp_data, conv_table_path)
    exp_data = preprocess.zscore(exp_data)
    bkgd = preprocess.background_df(exp_data)

    overexp_members = biclusters.make_membership_dictionary(reference_modules, bkgd, label=2, p=0.1)
    overexp_members_matrix = biclusters.membership_to_incidence(overexp_members, exp_data)

    underexp_members = biclusters.make_membership_dictionary(reference_modules, bkgd, label=0, p=0.1)
    underexp_members_matrix = biclusters.membership_to_incidence(underexp_members, exp_data)
    return overexp_members_matrix, underexp_members_matrix


def get_survival_subset(surv_df, label):
    sub_surv = surv_df[surv_df.index==label]
    sub_surv.index = sub_surv.iloc[:,0]
    sub_surv_df = sub_surv.loc[:,["D_PFS","D_PFS_FLAG"]]
    sub_surv_df.columns = ["duration","observed"]

    km_df = survival.km_analysis(survivalDf=sub_surv_df,
                                 durationCol="duration", statusCol="observed")
    return survival.guan_rank(kmSurvival=km_df)


def read_datasets(input_spec, verbose):
    """
    Reads a datasets dictionary from the specified input specification
    """
    with open(input_spec['regulons']) as infile:
        regulon_modules = json.load(infile)

    prim_survival = pd.read_csv(input_spec['primary_survival'], index_col=0, header=0)
    prim_survival_df = prim_survival.iloc[:,0:2]
    prim_survival_df.columns = ["duration", "observed"]

    test_survival = pd.read_csv(input_spec['test_survival'], index_col=0,header=0)

    datasets = []
    for i, dataset in enumerate(input_spec['datasets']):
        if verbose:
            logging.info("generating model for dataset '%s' (%d of %d)",
                         dataset['label'], i + 1, len(input_spec['datasets']))
        omm, umm = load_test_set(dataset['exp'], regulon_modules, dataset['idmap'])
        dataset_obj = {
            'omm': omm,
            'umm': umm,
            'label': dataset['label'],
            'primary': dataset['primary'],
            'dfr': (omm - umm)
        }
        if dataset['primary']:
            dataset_obj['km'] = survival.km_analysis(survivalDf=prim_survival_df,
                                                     durationCol="duration", statusCol="observed")
            dataset_obj['gs'] = survival.guan_rank(kmSurvival=dataset_obj['km'])
        else:
            dataset_obj['gs'] = get_survival_subset(test_survival, dataset['survival_subset'])

        datasets.append(dataset_obj)
    return datasets




from lifelines import KaplanMeierFitter
from scipy import stats
import numpy
import pandas
from lifelines import CoxPHFitter
import matplotlib.pyplot as plt
from miner2 import util


# =============================================================================
# Functions used for survival analysis
# =============================================================================

def km_analysis(survivalDf, durationCol, statusCol, saveFile=None):
    kmf = KaplanMeierFitter()
    kmf.fit(survivalDf.loc[:,durationCol],survivalDf.loc[:,statusCol])
    survFunc = kmf.survival_function_

    m, b, r, p, e = stats.linregress(list(survFunc.index),survFunc.iloc[:,0])

    survivalDf = survivalDf.sort_values(by=durationCol)
    ttpfs = numpy.array(survivalDf.loc[:,durationCol])
    survTime = numpy.array(survFunc.index)
    survProb = []

    for i in range(len(ttpfs)):
        date = ttpfs[i]
        if date in survTime:
            survProb.append(survFunc.loc[date,"KM_estimate"])
        elif date not in survTime:
            lbix = numpy.where(numpy.array(survFunc.index)<date)[0][-1]
            est = 0.5*(survFunc.iloc[lbix,0]+survFunc.iloc[lbix+1,0])
            survProb.append(est)

    kmEstimate = pandas.DataFrame(survProb)
    kmEstimate.columns = ["kmEstimate"]
    kmEstimate.index = survivalDf.index

    pfsDf = pandas.concat([survivalDf,kmEstimate],axis=1)

    if saveFile is not None:
        pfsDf.to_csv(saveFile)

    return pfsDf


def guan_rank(kmSurvival, saveFile=None):
    gScore = []
    for A in range(kmSurvival.shape[0]):
        aScore = 0
        aPfs = kmSurvival.iloc[A,0]
        aStatus = kmSurvival.iloc[A,1]
        aProbPFS = kmSurvival.iloc[A,2]
        if aStatus == 1:
            for B in range(kmSurvival.shape[0]):
                if B == A:
                    continue
                bPfs = kmSurvival.iloc[B,0]
                bStatus = kmSurvival.iloc[B,1]
                bProbPFS = kmSurvival.iloc[B,2]
                if bPfs > aPfs:
                    aScore+=1
                if bPfs <= aPfs:
                    if bStatus == 0:
                        aScore+=aProbPFS/bProbPFS
                if bPfs == aPfs:
                    if bStatus == 1:
                        aScore+=0.5
        elif aStatus == 0:
            for B in range(kmSurvival.shape[0]):
                if B == A:
                    continue
                bPfs = kmSurvival.iloc[B,0]
                bStatus = kmSurvival.iloc[B,1]
                bProbPFS = kmSurvival.iloc[B,2]
                if bPfs >= aPfs:
                    if bStatus == 0:
                        tmp = 1-0.5*bProbPFS/aProbPFS
                        aScore+=tmp
                    elif bStatus == 1:
                        tmp = 1-bProbPFS/aProbPFS
                        aScore+=tmp
                if bPfs < aPfs:
                    if bStatus == 0:
                        aScore+=0.5*aProbPFS/bProbPFS
        gScore.append(aScore)

    GuanScore = pandas.DataFrame(gScore)
    GuanScore = GuanScore/float(max(gScore))
    GuanScore.index = kmSurvival.index
    GuanScore.columns = ["GuanScore"]
    survivalData = pandas.concat([kmSurvival,GuanScore],axis=1)
    survivalData.sort_values(by="GuanScore", ascending=False, inplace=True)

    if saveFile is not None:
        survivalData.to_csv(saveFile)

    return survivalData



def survival_membership_analysis(task):

    start, stop = task[0]
    membershipDf,SurvivalDf = task[1]

    overlapPatients = list(set(membershipDf.columns)&set(SurvivalDf.index))
    if len(overlapPatients) == 0:
        print("samples are not represented in the survival data")
        return
    Survival = SurvivalDf.loc[overlapPatients,SurvivalDf.columns[0:2]]

    coxResults = {}
    keys = membershipDf.index[start:stop]
    ct=0
    for key in keys:
        ct+=1
        if ct%10==0:
            print(ct)
        try:
            memberVector = pandas.DataFrame(membershipDf.loc[key,overlapPatients])
            Survival2 = pandas.concat([Survival,memberVector],axis=1)
            Survival2.sort_values(by=Survival2.columns[0],inplace=True)

            cph = CoxPHFitter()
            cph.fit(Survival2, duration_col=Survival2.columns[0], event_col=Survival2.columns[1])

            tmpcph = cph.summary

            cox_hr = tmpcph.loc[key,"z"]
            cox_p = tmpcph.loc[key,"p"]
            coxResults[key] = (cox_hr, cox_p)
        except:
            coxResults[key] = (0, 1)
    return coxResults


def survival_median_analysis(task):
    start, stop = task[0]
    referenceDictionary,expressionDf,SurvivalDf = task[1]

    overlapPatients = list(set(expressionDf.columns)&set(SurvivalDf.index))
    Survival = SurvivalDf.loc[overlapPatients,SurvivalDf.columns[0:2]]

    coxResults = {}
    keys = referenceDictionary.keys()[start:stop]
    ct=0
    for key in keys:
        ct+=1
        if ct%10==0:
            print(ct)
        try:
            geneset = referenceDictionary[key]
            cluster = expressionDf.loc[geneset,overlapPatients]
            nz = np.count_nonzero(cluster+4.01,axis=0)

            medians = []
            for i in range(cluster.shape[1]):
                if nz[i] >= 3:
                    median = np.median(cluster.iloc[:,i][cluster.iloc[:,i]>-4.01])
                elif nz[i] < 3:
                    median = np.median(cluster.iloc[:,i])
                medians.append(median)

            medianDf = pd.DataFrame(medians)
            medianDf.index = overlapPatients
            medianDf.columns = ["median"]
            Survival = pd.concat([Survival,medianDf],axis=1)
            Survival.sort_values(by=Survival.columns[0],inplace=True)

            cph = CoxPHFitter()
            cph.fit(Survival, duration_col=Survival.columns[0], event_col=Survival.columns[1])

            tmpcph = cph.summary

            cox_hr = tmpcph.loc[key,"z"]
            cox_p = tmpcph.loc[key,"p"]
            coxResults[key] = (cox_hr, cox_p)
        except:
            coxResults[key] = (0, 1)

    return coxResults


def parallel_member_survival_analysis(membershipDf, numCores=5, survivalPath=None,
                                      survivalData=None):

    if survivalData is None:
        survivalData = pandas.read_csv(survivalPath,index_col=0,header=0)

    taskSplit = util.split_for_multiprocessing(membershipDf.index, numCores)
    taskData = (membershipDf, survivalData)
    tasks = [[taskSplit[i],taskData] for i in range(len(taskSplit))]
    coxOutput = util.multiprocess(survival_membership_analysis, tasks)
    return util.condense_output(coxOutput)


def parallel_median_survival_analysis(referenceDictionary,
                                      expressionDf,
                                      numCores=5,
                                      survivalPath=None,
                                      survivalData=None):

    if survivalData is None:
        survivalData = pd.read_csv(survivalPath,index_col=0,header=0)
    taskSplit = util.split_for_multiprocessing(referenceDictionary.keys(),numCores)
    taskData = (referenceDictionary,expressionDf,survivalData)
    tasks = [[taskSplit[i],taskData] for i in range(len(taskSplit))]
    coxOutput = util.multiprocess(survival_median_analysis,tasks)
    return util.condense_output(coxOutput)


def combined_states(groups, ranked_groups, survivalDf, minSamples=4, maxStates=7):
    high_risk_indices = []
    for i in range(1, len(ranked_groups) + 1):
        tmp_group = ranked_groups[-i]
        tmp_len = len(set(survivalDf.index) & set(groups[tmp_group]))
        if tmp_len >= minSamples:
            high_risk_indices.append(tmp_group)
        if len(high_risk_indices) >= maxStates:
            break

    combinations_high = []
    for i in range(len(high_risk_indices)-1):
        combinations_high.append(high_risk_indices[0: i + 1])

    low_risk_indices = []
    for i in range(len(ranked_groups)):
        tmp_group = ranked_groups[i]
        tmp_len = len(set(survivalDf.index) & set(groups[tmp_group]))
        if tmp_len >= minSamples:
            low_risk_indices.append(tmp_group)
        if len(low_risk_indices) >= maxStates:
            break

    combinations_low = []
    for i in range(len(low_risk_indices) - 1):
        combinations_low.append(low_risk_indices[0: i + 1])

    combined_states_high = []
    for i in range(len(combinations_high)):
        tmp = []
        for j in range(len(combinations_high[i])):
            tmp.append(groups[combinations_high[i][j]])
        combined_states_high.append(numpy.hstack(tmp))

    combined_states_low = []
    for i in range(len(combinations_low)):
        tmp = []
        for j in range(len(combinations_low[i])):
            tmp.append(groups[combinations_low[i][j]])
        combined_states_low.append(numpy.hstack(tmp))

    combined_states = numpy.concatenate([combined_states_high, combined_states_low])
    combined_indices_high = ["&".join(numpy.array(combinations_high[i]).astype(str))
                             for i in range(len(combinations_high))]
    combined_indices_low = ["&".join(numpy.array(combinations_low[i]).astype(str))
                            for i in range(len(combinations_low))]
    combined_indices = numpy.concatenate([combined_indices_high, combined_indices_low])

    return combined_states, combined_indices


def kmplot(srv,groups,labels,xlim_=None,filename=None,color=None,lw=1):

    for group in groups:
        try:
            patients = list(set(srv.index)&set(group))
            kmDf = kmAnalysis(survivalDf=srv.loc[patients,["duration","observed"]],
                              durationCol="duration",statusCol="observed")
            subset = kmDf[kmDf.loc[:,"observed"] == 1]
            duration = numpy.concatenate([numpy.array([0]), numpy.array(subset.loc[:,"duration"])])
            kme = numpy.concatenate([numpy.array([1]), numpy.array(subset.loc[:,"kmEstimate"])])
            if color is not None:
                plt.step(duration, kme, color=color, LineWidth=lw)
            elif color is None:
                plt.step(duration, kme, LineWidth=lw)
        except:
            continue

    try:
        axes = plt.gca()
        axes.set_axis_bgcolor([1,1,1])
        axes.grid(False)
        axes.spines['bottom'].set_color('0.25')
        axes.spines['top'].set_color('0.25')
        axes.spines['right'].set_color('0.25')
        axes.spines['left'].set_color('0.25')
        #axes.set_xlim([xmin,xmax])
        axes.set_ylim([0,1.09])
        if xlim_ is not None:
            axes.set_xlim([xlim_[0],xlim_[1]])
        #axes.set_title("Progression-free survival",FontSize=16)
        axes.set_ylabel("Disease-free",FontSize=14)
        axes.set_xlabel("Days",FontSize=14)
        if labels is not None:
            axes.legend(labels = labels,fontsize='x-small',ncol=3,loc='upper right')
        if filename is not None:
            plt.savefig(filename,bbox_inches="tight")
    except:
        print('Could not complete Kaplan-Meier analysis and plotting')
